@@ -12,6 +12,14 @@ Communication with TDM.
 from tdmclient.fb import FlatBuffer, Table, Union
 from tdmclient.zeroconf import TDMZeroconfBrowser
 from tdmclient.tcp import TDMConnection
-from tdmclient.thymio import ThymioFB
+from tdmclient.thymio import ThymioFB, Node
 from tdmclient.client import Client
 from tdmclient.clientasync import ClientAsync
+from tdmclient.clientnode import ClientNode
+from tdmclient.clientasyncnode import ClientAsyncNode
+from tdmclient.clientasynccachenode import ClientAsyncCacheNode, ArrayCache
+
+from tdmclient.server import Server, ServerNode
+
+# shortcut
+aw = ClientAsync.aw
