@@ -14,7 +14,227 @@ Author: Yves Piguet, EPFL
 from tdmclient import FlatBuffer, Union, Table
 
 
-class ThymioFB:
+class Listener:
+    """Base functionality for objects which are notified of variable changes
+    and events.
+    """
+
+    def __init__(self):
+        # on_variables_changed(node, variable_dict)
+        self.on_variables_changed = set()
+        # on_events_received(node, event_dict)
+        self.on_events_received = set()
+        # on_event_received(node, event_name, event_data)
+        self.on_event_received = set()
+
+    def add_variables_changed_listener(self, listener):
+        self.on_variables_changed.add(listener)
+
+    def remove_variables_changed_listener(self, listener):
+        self.on_variables_changed.remove(listener)
+
+    def add_events_received_listener(self, listener):
+        self.on_events_received.add(listener)
+
+    def remove_events_received_listener(self, listener):
+        self.on_events_received.remove(listener)
+
+    def add_event_received_listener(self, listener):
+        self.on_event_received.add(listener)
+
+    def remove_event_received_listener(self, listener):
+        self.on_event_received.remove(listener)
+
+    def notify_variables_changed(self, node, variable_dict):
+        for f in self.on_variables_changed:
+            f(node, variable_dict)
+
+    def notify_events_received(self, node, event_dict):
+        for f in self.on_events_received:
+            f(node, event_dict)
+        for f in self.on_event_received:
+            for name in event_dict:
+                f(node, name, event_dict[name])
+
+
+class Node(Listener):
+
+    def __init__(self, thymio, properties):
+        super(Node, self).__init__()
+
+        self.thymio = thymio
+        self.set_properties(properties)
+        self.vm_description = None
+
+    def set_properties(self, properties):
+        self.props = properties
+        self.id = properties["node_id"]
+        self.id_str = properties["node_id_str"]
+        self.status = properties["status"]
+
+    def __repr__(self):
+        return f"Node {self.id_str}"
+
+    def create_msg_request_vm_description(self, **kwargs):
+        return ThymioFB.create_message((
+            ThymioFB.MESSAGE_TYPE_REQUEST_NODE_ASEBA_VM_DESCRIPTION,
+            (
+                self.thymio.next_request_id(**kwargs),
+                (
+                    self.id,
+                ),
+            ),
+
+        ), ThymioFB.SCHEMA)
+
+    def create_msg_lock_node(self, **kwargs):
+        return ThymioFB.create_message((
+            ThymioFB.MESSAGE_TYPE_LOCK_NODE,
+            (
+                self.thymio.next_request_id(**kwargs),
+                (
+                    self.id,
+                ),
+            )
+        ), ThymioFB.SCHEMA)
+
+    def create_msg_unlock_node(self, **kwargs):
+        return ThymioFB.create_message((
+            ThymioFB.MESSAGE_TYPE_UNLOCK_NODE,
+            (
+                self.thymio.next_request_id(**kwargs),
+                (
+                    self.id,
+                ),
+            )
+        ), ThymioFB.SCHEMA)
+
+    def create_msg_rename_node(self, name, **kwargs):
+        return ThymioFB.create_message((
+            ThymioFB.MESSAGE_TYPE_RENAME_NODE,
+            (
+                self.thymio.next_request_id(**kwargs),
+                (
+                    self.id,
+                ),
+                name,
+            )
+        ), ThymioFB.SCHEMA)
+
+    def create_msg_program(self, program, load=True, **kwargs):
+        return ThymioFB.create_message((
+            ThymioFB.MESSAGE_TYPE_COMPILE_AND_LOAD_CODE_ON_VM,
+            (
+                self.thymio.next_request_id(**kwargs),
+                (
+                    self.id,
+                ),
+                ThymioFB.PROGRAMMING_LANGUAGE_ASEBA,
+                program,
+                ThymioFB.COMPILATION_OPTION_LOAD_ON_TARGET
+                    if load
+                    else ThymioFB.COMPILATION_OPTION_NONE,
+            )
+        ), ThymioFB.SCHEMA)
+
+    def create_msg_scratchpad_update(self, program, **kwargs):
+        return ThymioFB.create_message((
+            ThymioFB.MESSAGE_TYPE_SCRATCHPAD_UPDATE,
+            (
+                self.thymio.next_request_id(**kwargs),
+                (
+                    self.id,
+                ),
+                None,
+                (
+                    self.id,
+                ),
+                ThymioFB.PROGRAMMING_LANGUAGE_ASEBA,
+                program,
+            )
+        ), ThymioFB.SCHEMA)
+
+    def create_msg_set_vm_execution_state(self, state, **kwargs):
+        return ThymioFB.create_message((
+            ThymioFB.MESSAGE_TYPE_SET_VM_EXECUTION_STATE,
+            (
+                self.thymio.next_request_id(**kwargs),
+                (
+                    self.id,
+                ),
+                state
+            )
+        ), ThymioFB.SCHEMA)
+
+    def create_msg_watch_node(self, flags, **kwargs):
+        return ThymioFB.create_message((
+            ThymioFB.MESSAGE_TYPE_WATCH_NODE,
+            (
+                self.thymio.next_request_id(**kwargs),
+                (
+                    self.id,
+                ),
+                flags
+            )
+        ), ThymioFB.SCHEMA)
+
+    def create_msg_register_events(self, events, **kwargs):
+        return ThymioFB.create_message((
+            ThymioFB.MESSAGE_TYPE_REGISTER_EVENTS,
+            (
+                self.thymio.next_request_id(**kwargs),
+                (
+                    self.id,
+                ),
+                [
+                    (
+                        event[0], # name (str)
+                        event[1], # fixed size (int)
+                        0, # index (int)
+                    )
+                    for event in events
+                ]
+            )
+        ), ThymioFB.SCHEMA)
+
+    def create_msg_send_events(self, event_dict, **kwargs):
+        return ThymioFB.create_message((
+            ThymioFB.MESSAGE_TYPE_SEND_EVENTS,
+            (
+                self.thymio.next_request_id(**kwargs),
+                (
+                    self.id,
+                ),
+                [
+                    (
+                        name,
+                        event_dict[name],
+                    )
+                    for name in event_dict
+                ]
+            )
+        ), ThymioFB.SCHEMA)
+
+    def create_msg_set_variables(self, var_dict, **kwargs):
+        return ThymioFB.create_message((
+            ThymioFB.MESSAGE_TYPE_SET_VARIABLES,
+            (
+                self.thymio.next_request_id(**kwargs),
+                (
+                    self.id,
+                ),
+                [
+                    (
+                        name,
+                        var_dict[name],
+                    )
+                    for name in var_dict
+                ]
+            )
+        ), ThymioFB.SCHEMA)
+
+
+class ThymioFB(Listener):
 
     MESSAGE_TYPE_CONNECTION_HANDSHAKE = 1
     MESSAGE_TYPE_DEVICE_MANAGER_SHUTDOWN_REQUEST = 2
@@ -107,6 +327,7 @@ class ThymioFB:
     VM_EXECUTION_STATE_PAUSED = 2
 
     def __init__(self, debug=0):
+        super(ThymioFB, self).__init__()
 
         self.debug = debug
 
@@ -116,18 +337,13 @@ class ThymioFB:
         self.last_request_id = 0
         self.request_id_notify_dict = {}
 
+        # on_nodes_changed(node_list)
         self.on_nodes_changed = None
-        self.on_variables_changed = None
 
-    def create_message(self, msg, schema=None):
-        fb = FlatBuffer()
-        if schema is None:
-            fb.load_from_native_type(msg)
-        else:
-            fb.load_with_schema(msg, schema)
-        encoded_fb = fb.encode()
-
-        return encoded_fb
+    def create_node(self, node_dict):
+        """Create a Node object, of class Node or a subclass.
+        """
+        return Node(self, node_dict)
 
     def next_request_id(self, request_id_notify=None):
         self.last_request_id += 1
@@ -135,12 +351,6 @@ class ThymioFB:
             self.request_id_notify_dict[self.last_request_id] = request_id_notify
         return self.last_request_id
 
-    def get_node_id(self, node_id_str):
-        for node in self.nodes:
-            if node["node_id_str"] == node_id_str:
-                return node["node_id"]
-        raise Exception("node id not found")
-
     SCHEMA = """
         // see https://github.com/Mobsya/aseba/blob/master/aseba/flatbuffers/thymio.fbs
         // root object: union AnyMessage
@@ -238,175 +448,192 @@ class ThymioFB:
         )
     """
 
+    @staticmethod
+    def create_message(msg, schema=None):
+        fb = FlatBuffer()
+        if schema is None:
+            fb.load_from_native_type(msg)
+        else:
+            fb.load_with_schema(msg, schema)
+        encoded_fb = fb.encode()
+
+        return encoded_fb
+
     def create_msg_handshake(self):
         return self.create_message((
-            self.MESSAGE_TYPE_CONNECTION_HANDSHAKE,
+            ThymioFB.MESSAGE_TYPE_CONNECTION_HANDSHAKE,
             ()
-        ), self.SCHEMA)
+        ), ThymioFB.SCHEMA)
 
-    def create_msg_lock_node(self, node_id_str, **kwargs):
+    def create_msg_device_manager_shutdown_request(self, **kwargs):
         return self.create_message((
-            self.MESSAGE_TYPE_LOCK_NODE,
+            ThymioFB.MESSAGE_TYPE_DEVICE_MANAGER_SHUTDOWN_REQUEST,
             (
                 self.next_request_id(**kwargs),
-                (
-                    self.get_node_id(node_id_str),
-                ),
             )
-        ), self.SCHEMA)
+        ), ThymioFB.SCHEMA)
 
-    def create_msg_unlock_node(self, node_id_str, **kwargs):
+    def create_msg_request_list_of_nodes(self):
         return self.create_message((
-            self.MESSAGE_TYPE_UNLOCK_NODE,
-            (
-                self.next_request_id(**kwargs),
-                (
-                    self.get_node_id(node_id_str),
-                ),
-            )
-        ), self.SCHEMA)
+            ThymioFB.MESSAGE_TYPE_REQUEST_LIST_OF_NODES,
+        ), ThymioFB.SCHEMA)
 
-    def create_msg_program(self, node_id_str, program, load=True, **kwargs):
+    def create_msg_request_completed(self, request_id):
         return self.create_message((
-            self.MESSAGE_TYPE_COMPILE_AND_LOAD_CODE_ON_VM,
+            ThymioFB.MESSAGE_TYPE_REQUEST_COMPLETED,
             (
-                self.next_request_id(**kwargs),
-                (
-                    self.get_node_id(node_id_str),
-                ),
-                self.PROGRAMMING_LANGUAGE_ASEBA,
-                program,
-                self.COMPILATION_OPTION_LOAD_ON_TARGET
-                    if load
-                    else self.COMPILATION_OPTION_NONE,
+                request_id,
             )
-        ), self.SCHEMA)
+        ), ThymioFB.SCHEMA)
 
-    def create_msg_scratchpad_update(self, node_id_str, program, **kwargs):
+    def create_msg_error(self, request_id, error):
         return self.create_message((
-            self.MESSAGE_TYPE_SCRATCHPAD_UPDATE,
+            ThymioFB.MESSAGE_TYPE_ERROR,
             (
-                self.next_request_id(**kwargs),
-                (
-                    self.get_node_id(node_id_str),
-                ),
-                None,
-                (
-                    self.get_node_id(node_id_str),
-                ),
-                self.PROGRAMMING_LANGUAGE_ASEBA,
-                program,
+                request_id,
+                error,
             )
-        ), self.SCHEMA)
+        ), ThymioFB.SCHEMA)
 
-    def create_msg_set_vm_execution_state(self, node_id_str, state, **kwargs):
+    def create_msg_compilation_result_failure(self, request_id, message, character, line, column):
         return self.create_message((
-            self.MESSAGE_TYPE_SET_VM_EXECUTION_STATE,
+            ThymioFB.MESSAGE_TYPE_COMPILATION_RESULT_FAILURE,
             (
-                self.next_request_id(**kwargs),
-                (
-                    self.get_node_id(node_id_str),
-                ),
-                state
+                request_id,
+                message,
+                character,
+                line,
+                column,
             )
-        ), self.SCHEMA)
+        ), ThymioFB.SCHEMA)
 
-    def create_msg_watch_node(self, node_id_str, flags, **kwargs):
+    def create_msg_compilation_result_success(self, request_id, bc_size, total_bc_size, var_size, total_var_size):
         return self.create_message((
-            self.MESSAGE_TYPE_WATCH_NODE,
+            ThymioFB.MESSAGE_TYPE_COMPILATION_RESULT_SUCCESS,
             (
-                self.next_request_id(**kwargs),
-                (
-                    self.get_node_id(node_id_str),
-                ),
-                flags
+                request_id,
+                bc_size,
+                total_bc_size,
+                var_size,
+                total_var_size,
             )
-        ), self.SCHEMA)
+        ), ThymioFB.SCHEMA)
 
-    def create_msg_register_events(self, node_id_str, events, **kwargs):
-        return self.create_message((
-            self.MESSAGE_TYPE_REGISTER_EVENTS,
+    def create_msg_vm_execution_state_changed(self, node_id_str, state, line, error, error_msg):
+        return ThymioFB.create_message((
+            ThymioFB.MESSAGE_TYPE_VM_EXECUTION_STATE_CHANGED,
             (
-                self.next_request_id(**kwargs),
                 (
-                    self.get_node_id(node_id_str),
+                    ThymioFB.id_str_to_bytes(node_id_str),
                 ),
-                [
-                    (
-                        event[0], # name (str)
-                        event[1], # fixed size (int)
-                        0, # index (int)
-                    )
-                    for event in events
-                ]
+                state,
+                line,
+                error,
+                error_msg,
             )
-        ), self.SCHEMA)
+        ), ThymioFB.SCHEMA)
 
-    def create_msg_set_variables(self, node_id_str, var_dict, **kwargs):
-        return self.create_message((
-            self.MESSAGE_TYPE_SET_VARIABLES,
-            (
-                self.next_request_id(**kwargs),
-                (
-                    self.get_node_id(node_id_str),
-                ),
-                [
-                    (
-                        name,
-                        var_dict[name],
-                    )
-                    for name in var_dict
-                ]
-            )
-        ), self.SCHEMA)
+    def find_node(self, node_id_str):
+        for node in self.nodes:
+            if node.id_str == node_id_str:
+                return node
+
+    @staticmethod
+    def bytes_to_id_str(f):
+        if f is None:
+            return None
+        else:
+            str = "".join([f"{b if type(b) is int else ord(b):02x}" for b in f[0].fields[0][0]])
+            str = str[:8] + "-" + str[8:12] + "-" + str[12:16] + "-" + str[16:20] + "-" + str[20:]
+            return str
+
+    @staticmethod
+    def id_str_to_bytes(id_str):
+        id_str = id_str.replace("-", "")
+        b = []
+        for i in range(0, len(id_str), 2):
+            b.append(int(id_str[i : i + 2], 16))
+        return bytes(b)
 
     def process_message(self, msg):
 
-        def field_val(f, default):
-            return f[0] if f is not None else default
-
-        def bytes_to_hexa(f):
-            if f is None:
-                return None
-            else:
-                return "".join([f"{b if type(b) is int else ord(b):02x}" for b in f[0].fields[0][0]])
-
         fb = FlatBuffer()
-        fb.parse(msg, self.SCHEMA)
+        fb.parse(msg, ThymioFB.SCHEMA)
         if self.debug >= 3:
             fb.dump()
         if type(fb.root) is Union:
             if fb.root.union_type == self.MESSAGE_TYPE_PING:
                 pass
             elif fb.root.union_type == self.MESSAGE_TYPE_CONNECTION_HANDSHAKE:
-                self.protocol_version = field_val(fb.root.union_data[0].fields[1], 1)
-                self.localhost_peer = field_val(fb.root.union_data[0].fields[4], False)
+                self.protocol_version = FlatBuffer.field_val(fb.root.union_data[0].fields[1], 1)
+                self.localhost_peer = FlatBuffer.field_val(fb.root.union_data[0].fields[4], False)
             elif fb.root.union_type == self.MESSAGE_TYPE_NODES_CHANGED:
                 if fb.root.union_data[0] is not None:
                     nodes = fb.root.union_data[0].fields[0][0]
-                    self.nodes = [
-                        {
+                    for node in nodes:
+                        node_id_str = ThymioFB.bytes_to_id_str(node.fields[0])
+                        node_properties = {
                             "node_id":
                                 None if node.fields[0] is None
                                 else node.fields[0][0].fields[0][0] if type(node.fields[0][0].fields[0][0]) is bytes
                                 else b"".join(node.fields[0][0].fields[0][0]),
-                            "node_id_str": bytes_to_hexa(node.fields[0]),
-                            "group_id": field_val(node.fields[1], None),
-                            "status": field_val(node.fields[2], -1),
-                            "type": field_val(node.fields[3], -1),
-                            "name": field_val(node.fields[4], ""),
-                            "capabilities": field_val(node.fields[5], 0),
-                            "fw_version": field_val(node.fields[6], None),
+                            "node_id_str": ThymioFB.bytes_to_id_str(node.fields[0]),
+                            "group_id": FlatBuffer.field_val(node.fields[1], None),
+                            "status": FlatBuffer.field_val(node.fields[2], -1),
+                            "type": FlatBuffer.field_val(node.fields[3], -1),
+                            "name": FlatBuffer.field_val(node.fields[4], ""),
+                            "capabilities": FlatBuffer.field_val(node.fields[5], 0),
+                            "fw_version": FlatBuffer.field_val(node.fields[6], None),
                         }
-                        for node in nodes
-                    ]
+                        existing_node = self.find_node(node_id_str)
+                        if existing_node is not None:
+                            existing_node.set_properties(node_properties)
+                        else:
+                            self.nodes.append(self.create_node(node_properties))
                     if self.on_nodes_changed is not None:
                         self.on_nodes_changed(self.nodes)
                     if self.debug >= 1:
                         print("NodesChanged",
-                              ", ".join(f"{node['node_id_str']}: status={node['status']}" for node in self.nodes))
+                              ", ".join(f"{node.id_str}: status={node.status}" for node in self.nodes))
+            elif fb.root.union_type == self.MESSAGE_TYPE_NODE_ASEBA_VM_DESCRIPTION:
+                request_id = FlatBuffer.field_val(fb.root.union_data[0].fields[0], 0)
+                vm_descr = fb.root.union_data[0]
+                node_id_str = ThymioFB.bytes_to_id_str(vm_descr.fields[1])
+                node = self.find_node(node_id_str)
+                vm_description = {
+                    "node_id":
+                        None if vm_descr.fields[1] is None
+                        else vm_descr.fields[1][0].fields[0][0] if type(vm_descr.fields[1][0].fields[0][0]) is bytes
+                        else b"".join(vm_descr.fields[1][0].fields[0][0]),
+                    "node_id_str": node_id_str,
+                    "bytecode_size": FlatBuffer.field_val(vm_descr.fields[2], None),
+                    "data_size": FlatBuffer.field_val(vm_descr.fields[3], None),
+                    "stack_size": FlatBuffer.field_val(vm_descr.fields[4], None),
+                    "variables": {
+                        FlatBuffer.field_val(var_descr.fields[1], ""):
+                            (lambda s: None if s == 1 else s)(FlatBuffer.field_val(var_descr.fields[2], None))
+                        for var_descr in vm_descr.fields[5][0]
+                    },
+                    "events": [
+                        FlatBuffer.field_val(event_descr.fields[1], "")
+                        for event_descr in vm_descr.fields[6][0]
+                    ],
+                    "functions": {
+                        FlatBuffer.field_val(fun_descr.fields[1], ""): [
+                            FlatBuffer.field_val(param.fields[1], None)
+                            for param in fun_descr.fields[3][0]
+                        ]
+                        for fun_descr in vm_descr.fields[7][0]
+                    },
+                }
+                if node is not None:
+                    node.vm_description = vm_description
+                if request_id in self.request_id_notify_dict:
+                    self.request_id_notify_dict[request_id](vm_description)
+                elif self.debug >= 1:
+                    print(f"vm description request_id={request_id} (ignored)")
             elif fb.root.union_type == self.MESSAGE_TYPE_REQUEST_COMPLETED:
-                request_id = field_val(fb.root.union_data[0].fields[0], 0)
+                request_id = FlatBuffer.field_val(fb.root.union_data[0].fields[0], 0)
                 if request_id in self.request_id_notify_dict:
                     self.request_id_notify_dict[request_id](None)
                     del self.request_id_notify_dict[request_id]
@@ -415,8 +642,8 @@ class ThymioFB:
                 elif self.debug >= 1:
                     print(f"ok request_id={request_id} (ignored)")
             elif fb.root.union_type == self.MESSAGE_TYPE_ERROR:
-                request_id = field_val(fb.root.union_data[0].fields[0], 0)
-                error_code = field_val(fb.root.union_data[0].fields[1], 0)
+                request_id = FlatBuffer.field_val(fb.root.union_data[0].fields[0], 0)
+                error_code = FlatBuffer.field_val(fb.root.union_data[0].fields[1], 0)
                 if request_id in self.request_id_notify_dict:
                     self.request_id_notify_dict[request_id]({"error_code": error_code})
                     del self.request_id_notify_dict[request_id]
@@ -425,10 +652,10 @@ class ThymioFB:
                 elif self.debug >= 1:
                     print(f"error {error_code} request_id={request_id} (ignored)")
             elif fb.root.union_type == self.MESSAGE_TYPE_COMPILATION_RESULT_FAILURE:
-                request_id = field_val(fb.root.union_data[0].fields[0], 0)
-                error_msg = field_val(fb.root.union_data[0].fields[1], "")
-                error_line = field_val(fb.root.union_data[0].fields[3], 0)
-                error_col = field_val(fb.root.union_data[0].fields[4], 0)
+                request_id = FlatBuffer.field_val(fb.root.union_data[0].fields[0], 0)
+                error_msg = FlatBuffer.field_val(fb.root.union_data[0].fields[1], "")
+                error_line = FlatBuffer.field_val(fb.root.union_data[0].fields[3], 0)
+                error_col = FlatBuffer.field_val(fb.root.union_data[0].fields[4], 0)
                 if request_id in self.request_id_notify_dict:
                     self.request_id_notify_dict[request_id]({
                         "error_msg": error_msg,
@@ -441,7 +668,7 @@ class ThymioFB:
                 elif self.debug >= 1:
                     print(f"compilation error: {error_msg} request_id={request_id} (ignored)")
             elif fb.root.union_type == self.MESSAGE_TYPE_COMPILATION_RESULT_SUCCESS:
-                request_id = field_val(fb.root.union_data[0].fields[0], 0)
+                request_id = FlatBuffer.field_val(fb.root.union_data[0].fields[0], 0)
                 if request_id in self.request_id_notify_dict:
                     self.request_id_notify_dict[request_id](None)
                     del self.request_id_notify_dict[request_id]
@@ -450,27 +677,29 @@ class ThymioFB:
                 elif self.debug >= 1:
                     print(f"compilation ok request_id={request_id} (ignored)")
             elif fb.root.union_type == self.MESSAGE_TYPE_VARIABLES_CHANGED:
-                node_id_str = bytes_to_hexa(fb.root.union_data[0].fields[0])
+                node_id_str = ThymioFB.bytes_to_id_str(fb.root.union_data[0].fields[0])
+                node = self.find_node(node_id_str)
                 variables = {
                     v.fields[0][0]: v.fields[1][0]
                     for v in fb.root.union_data[0].fields[1][0]
                 }
-                if self.on_variables_changed is not None:
-                    self.on_variables_changed(node_id_str, {"variables": variables})
+                self.notify_variables_changed(node, variables)
+                if node is not None:
+                    node.notify_variables_changed(node, variables)
                 if self.debug >= 1:
                     print(f"variables of node {node_id_str} changed")
                     if self.debug >= 2:
                         for name in variables:
                             print(name, variables[name])
             elif fb.root.union_type == self.MESSAGE_TYPE_EVENTS_DESCRIPTIONS_CHANGED:
-                node_or_group_id = bytes_to_hexa(fb.root.union_data[0].fields[0])
+                node_or_group_id = ThymioFB.bytes_to_id_str(fb.root.union_data[0].fields[0])
                 if fb.root.union_data[0].fields[1] is not None:
                     event_size = {
-                        e.fields[0][0]: field_val(e.fields[1], 0)
+                        e.fields[0][0]: FlatBuffer.field_val(e.fields[1], 0)
                         for e in fb.root.union_data[0].fields[1][0]
                     }
                     event_index = {
-                        e.fields[0][0]: field_val(e.fields[2], 0)
+                        e.fields[0][0]: FlatBuffer.field_val(e.fields[2], 0)
                         for e in fb.root.union_data[0].fields[1][0]
                     }
                 else:
@@ -483,11 +712,14 @@ class ThymioFB:
                             f"{name}[{event_size[name]}]" for name in event_size
                         ]))
             elif fb.root.union_type == self.MESSAGE_TYPE_EVENTS_EMITTED:
-                node_id_str = bytes_to_hexa(fb.root.union_data[0].fields[0])
+                node_id_str = ThymioFB.bytes_to_id_str(fb.root.union_data[0].fields[0])
+                node = self.find_node(node_id_str)
                 events = {
                     e.fields[0][0]: e.fields[1][0]
                     for e in fb.root.union_data[0].fields[1][0]
                 }
+                self.notify_events_received(node, events)
+                node.notify_events_received(node, events)
                 if self.debug >= 1:
                     print(f"events emitted by node {node_id_str}")
                     if self.debug >= 2:
@@ -495,22 +727,22 @@ class ThymioFB:
                             print(name,
                                   events[name] if events[name] is not None else "")
             elif fb.root.union_type == self.MESSAGE_TYPE_VM_EXECUTION_STATE_CHANGED:
-                node_id_str = bytes_to_hexa(fb.root.union_data[0].fields[0])
-                state = field_val(fb.root.union_data[0].fields[1], 0)
-                line = field_val(fb.root.union_data[0].fields[2], 0)
-                error = field_val(fb.root.union_data[0].fields[3], 0)
-                error_msg = field_val(fb.root.union_data[0].fields[4], "")
+                node_id_str = ThymioFB.bytes_to_id_str(fb.root.union_data[0].fields[0])
+                state = FlatBuffer.field_val(fb.root.union_data[0].fields[1], 0)
+                line = FlatBuffer.field_val(fb.root.union_data[0].fields[2], 0)
+                error = FlatBuffer.field_val(fb.root.union_data[0].fields[3], 0)
+                error_msg = FlatBuffer.field_val(fb.root.union_data[0].fields[4], "")
                 if self.debug >= 1:
                     print(f"execution state of node {node_id_str} changed to {state}, line={line}, error={error} {error_msg}")
             elif fb.root.union_type == self.MESSAGE_TYPE_SCRATCHPAD_UPDATE:
-                request_id = field_val(fb.root.union_data[0].fields[0], 0)
-                scratchpad_id_str = bytes_to_hexa(fb.root.union_data[0].fields[1])
-                group_id_str = bytes_to_hexa(fb.root.union_data[0].fields[2])
-                node_id_str = bytes_to_hexa(fb.root.union_data[0].fields[3])
-                language = field_val(fb.root.union_data[0].fields[4], self.PROGRAMMING_LANGUAGE_ASEBA)
-                text = field_val(fb.root.union_data[0].fields[5], "")
-                name = field_val(fb.root.union_data[0].fields[6], "")
-                deleted = field_val(fb.root.union_data[0].fields[7], False)
+                request_id = FlatBuffer.field_val(fb.root.union_data[0].fields[0], 0)
+                scratchpad_id_str = ThymioFB.bytes_to_id_str(fb.root.union_data[0].fields[1])
+                group_id_str = ThymioFB.bytes_to_id_str(fb.root.union_data[0].fields[2])
+                node_id_str = ThymioFB.bytes_to_id_str(fb.root.union_data[0].fields[3])
+                language = FlatBuffer.field_val(fb.root.union_data[0].fields[4], self.PROGRAMMING_LANGUAGE_ASEBA)
+                text = FlatBuffer.field_val(fb.root.union_data[0].fields[5], "")
+                name = FlatBuffer.field_val(fb.root.union_data[0].fields[6], "")
+                deleted = FlatBuffer.field_val(fb.root.union_data[0].fields[7], False)
                 if self.debug >= 1:
                     print(f"scratchpad {scratchpad_id_str} of node {node_id_str} / group {group_id_str} updated")
                     print(text)
